/*
	HEADER:		CUG267;
	TITLE:	8085 Cross-Assembler (Portable);
	FILENAME:	A85UTIL.C;
	VERSION:	0.1;
	DATE:		08/27/1988;
	SEE-ALSO:	A85.H;
	AUTHORS:	William C. Colley III;
*/

/*
		      8085 Cross-Assembler in Portable C

		Copyright (c) 1985,1987 William C. Colley, III

Revision History:

Ver	Date		Description

0.0	AUG 1987	Derived from version 3.4 of my portable 6800/6801
			cross-assembler.  WCC3.

0.1	AUG 1988	Fixed a bug in the command line parser that puts it
			into a VERY long loop if the user types a command line
			like "A85 FILE.ASM -L".  WCC3 per Alex Cameron.


	DEC 2013	changes by Herb Johnson to compile under lcc-32 for MS-DOS.
			fixes as late as April 2010. Also see "a18octal.txt" to
			change listing from hexidecimal values to octal.

	2013  	Fixes for Turbo C HRJ include line[] changed to lline[]

	Nov 2014	adapted for 4040 4004 assembler HRJ

0.2 Oct 2019	changes inspired by Dwight Elvey HRJ
0.21 Jul 2020	register pairs Elvey and HRJ



This module contains the following utility packages:

	1)  symbol table building and searching

	2)  opcode and operator table searching

	3)  listing file output

	4)  hex file output

	5)  error flagging
*/

/*  Get global goodies:  */

#include "a04.h"
#include <string.h> /* HRJ */
#include <ctype.h>
// #include <malloc.h> /* for lcc-32 HRJ */
/* #include <alloc.h> for Turbo C HRJ */
#include <stdlib.h>

/*  Make sure that MSDOS compilers using the large memory model know	*/
/*  that calloc() returns pointer to char as an MSDOS far pointer is	*/
/*  NOT compatible with the int type as is usually the case.		*/

/* char *calloc(); HRJ */

/*HRJ local declarations */

static OPCODE *bccsearch(OPCODE *, OPCODE *, char *);
static void list_sym(SYMBOL *);
static void record(unsigned);
static void putb(unsigned);
static int ustrcmp(char *, char*);
static void check_page(void);
void warning(char *);
void fatal_error(char *);


/*  Get access to global mailboxes defined in A04.C:			*/

extern char errcode, lline[], title[];
extern int eject, listhex;
extern unsigned address, bytes, errors, listleft, obj[], pagelen;

/*  The symbol table is a binary tree of variable-length blocks drawn	*/
/*  from the heap with the calloc() function.  The root pointer lives	*/
/*  here:								*/

static SYMBOL *sroot = NULL;

/*  Add new symbol to symbol table.  Returns pointer to symbol even if	*/
/*  the symbol already exists.  If there's not enough memory to store	*/
/*  the new symbol, a fatal error occurs.				*/

SYMBOL *new_symbol(char *nam)

{
    SCRATCH int i;
    SCRATCH SYMBOL **p, *q;
    void fatal_error(char *);

    /* printf("new_symbol>>%s<<\n",nam);  HRJ diagnostic*/

    for (p = &sroot; (q = *p) && (i = strcmp(nam,q -> sname)); )
	p = i < 0 ? &(q -> left) : &(q -> right);
    if (!q) {
	if (!(*p = q = (SYMBOL *)calloc(1,sizeof(SYMBOL) + strlen(nam))))
	    fatal_error(SYMBOLS);
	strcpy(q -> sname,nam);
    }
    return q;
}

/*  Look up symbol in symbol table.  Returns pointer to symbol or NULL	*/
/*  if symbol not found.						*/

SYMBOL *find_symbol(char *nam)

{
    SCRATCH int i;
    SCRATCH SYMBOL *p;

    for (p = sroot; p && (i = strcmp(nam,p -> sname));
	p = i < 0 ? p -> left : p -> right);
    return p;
}

/*  Opcode table search routine.  This routine pats down the opcode	*/
/*  table for a given opcode and returns either a pointer to it or	*/
/*  NULL if the opcode doesn't exist.					*/

OPCODE *find_code(char *nam)

{
    /* OPCODE *bsearch(); */

    static OPCODE opctbl[] = {

	{ NONE + 1,				0x00,	"NOP"	},

	{ NONE + 1,				0x01,	"HLT"	}, //HRJ 4040 only, 1 byte
	{ NONE + 1,				0x02,	"BBS"	},
	{ NONE + 1,				0x03,	"LCR"	},
	{ NONE + 1,				0x04,	"OR4"	},
	{ NONE + 1,				0x05,	"OR5"	},
	{ NONE + 1,				0x06,	"AN6"	},
	{ NONE + 1,				0x07,	"AN7"	},
	{ NONE + 1,				0x08,	"DB0"	},
	{ NONE + 1,				0x09,	"DB1"	},
	{ NONE + 1,				0x0A,	"SB0"	},
	{ NONE + 1,				0x0B,	"SB1"	},
	{ NONE + 1,				0x0C,	"EIN"	},
	{ NONE + 1,				0x0D,	"DIN"	},
	{ NONE + 1,				0x0E,	"RPM"	},
						// 0X0F undefined

	// 1x	JCN	condition, byte address
	// 0001cccc aaaaaaaa
	// can encode with DST_REG and ADDR_8
	// 	can also support some Intel condition codes

	{ DST_REG  + (ADDR_8 << 4) + 2, 0x10, "JCN" },
	{ ADDR_8 + 2, 			0x11, "JNT" },
	{ ADDR_8 + 2, 			0x12, "JC" },
	{ ADDR_8 + 2, 			0x14, "JZ" },
	{ ADDR_8 + 2, 			0x19, "JT" },
	{ ADDR_8 + 2, 			0x1A, "JNC" },
	{ ADDR_8 + 2, 			0x1C, "JNZ" },

	// 2x	FIM	reg pair, byte data
	// 0010rrr0 dddddddd
	// DEST_REGP  + (DATA_8 << 4) + 2 will encode
	// note: FIM may load a portion of an address that's used later
	// so relax requirement that address be within current ROM HRJ

	{ DST_REGP  + (DATA_8 << 4) + 2, 0x20, "FIM" },

	// 2x	SRC 	reg pair 0010rrr1 - for DRC selected RAM or I/O
	{ DST_REGP  + 1,		 0x21, "SRC" },


	// 3x	FIN	reg pair 0011rrr0
	{ DST_REGP  + 1,		 0x30, "FIN" },

	// 3x	JIN	reg pair 0010rrr1
	{ DST_REGP  + 1,		 0x31, "JIN" },

	// 4x	JUN	12-bit address
	// 0100aaaa aaaaaaaa

	{ DATA_12 + 2,			0x40, "JUN" },

	// 5x	JMS	12-bit address
	// 0100aaaa aaaaaaaa

	{ DATA_12 + 2,			0x50, "JMS" },

	// 6x	INC	reg 0110rrrr
	// DST_REG  will encode
	{ DST_REG + 1,			0x60,	"INC"	},

	// 7x	ISZ	reg, byte address 0111rrrr aaaaaaaa
	// DST_REG  + (ADDR_8 << 4)  will encode

	{ DST_REG + (ADDR_8 << 4) + 2, 0x70, "ISZ" },

	// 8x	ADD	reg 1000rrrr
	// DST_REG  will encode
	{ DST_REG + 1,			0x80,	"ADD"	},

	// 9X	SUB	reg 1001rrrr
	// DST_REG  will encode
	{ DST_REG + 1,			0x90,	"SUB"	},

	// Ax	LD	reg 1010rrrr
	// DST_REG  will encode
	{ DST_REG + 1,			0xA0,	"LD"	},

	// Bx	XCH	reg 1010rrrr
	// DST_REG  will encode
	{ DST_REG + 1,			0xB0,	"XCH"	},

	// Cx	BBL	4-bit data 1100dddd
	// DST_REG  will encode
	{ DST_REG + 1,			0xC0,	"BBL"	},

	// Dx	LDM	4-bit data 1101dddd
	// DET_REG  will encode
	{ DST_REG + 1,			0xD0,	"LDM"	},


	{ NONE + 1,				0xE0,	"WRM"	}, //HRJ I-O & RAM insructions
	{ NONE + 1,				0xE1,	"WMP"	},
	{ NONE + 1,				0xE2,	"WRR"	},
	{ NONE + 1,				0xE3,	"WPM"	},
	{ NONE + 1,				0xE4,	"WR0"	},
	{ NONE + 1,				0xE5,	"WR1"	},
	{ NONE + 1,				0xE6,	"WR2"	},
	{ NONE + 1,				0xE7,	"WR3"	},
	{ NONE + 1,				0xE8,	"SBM"	},
	{ NONE + 1,				0xE9,	"RDM"	},
	{ NONE + 1,				0xEA,	"RDR"	},
	{ NONE + 1,				0xEB,	"ADM"	},
	{ NONE + 1,				0xEC,	"RD0"	},
	{ NONE + 1,				0xED,	"RD1"	},
	{ NONE + 1,				0xEE,	"RD2"	},
	{ NONE + 1,				0xEF,	"RD3"	},

	{ NONE + 1,				0xF0,	"CLB"	}, //HRJ 4004
	{ NONE + 1,				0xF1,	"CLC"	},
	{ NONE + 1,				0xF2,	"IAC"	},
	{ NONE + 1,				0xF3,	"CMC"	},
	{ NONE + 1,				0xF4,	"CMA"	},
	{ NONE + 1,				0xF5,	"RAL"	},
	{ NONE + 1,				0xF6,	"RAR"	},
	{ NONE + 1,				0xF7,	"TCC"	},
	{ NONE + 1,				0xF8,	"DAC"	},
	{ NONE + 1,				0xF9,	"TCS"	},
	{ NONE + 1,				0xFA,	"STC"	},
	{ NONE + 1,				0xFB,	"DAA"	},
	{ NONE + 1,				0xFC,	"KBP"	},

	// DCL selects RAM -I/O bank from 3 LSB's of accumulator
	{ NONE + 1,				0xFD,	"DCL"	},

	//0xFE, 0xFF undefined


	{ PSEUDO,				DB,	"DB"	},
	{ PSEUDO,				DS,	"DS"	},
	{ PSEUDO,				DW,	"DW"	},
	{ PSEUDO + ISIF,		ELSE,	"ELSE"	},
	{ PSEUDO,				END,	"END"	},
	{ PSEUDO + ISIF,		ENDIF,	"ENDIF"	},
	{ PSEUDO,				EQU,	"EQU"	},
	{ PSEUDO + ISIF,		IF,	"IF"	},
	{ PSEUDO,				INCL,	"INCL"	},
	{ PSEUDO,				ORG,	"ORG"	},
	{ PSEUDO,				PAGE,	"PAGE"	},
	{ PSEUDO,				RPAIRS,	"RPAIRS" },
	{ PSEUDO,				SET,	"SET"	},
	{ PSEUDO,				TITLE,	"TITLE"	},

    };

    return bccsearch(opctbl,opctbl + (sizeof(opctbl) / sizeof(OPCODE)),nam);
}

/*  Operator table search routine.  This routine pats down the		*/
/*  operator table for a given operator and returns either a pointer	*/
/*  to it or NULL if the opcode doesn't exist.				*/

OPCODE *find_operator(char *nam)

{
    /* OPCODE *bsearch(); */

    static OPCODE oprtbl[] = {
	//{ BCDEHLMA + REG,				A,	"A"	},
	{ BINARY + LOG1  + OPR,				AND,	"AND"	},
	//{ BCDEHLMA + BDHPSW + BDHSP + BD + REG,	B,	"B"	},
	//{ BCDEHLMA + REG,				C,	"C"	},
	//{ BCDEHLMA + BDHPSW + BDHSP + BD + REG,	D,	"D"	},
	//{ BCDEHLMA + REG,				E,	"E"	},
	{ BINARY + RELAT + OPR,				'=',	"EQ"	},
	{ BINARY + RELAT + OPR,				GE,	"GE"	},
	{ BINARY + RELAT + OPR,				'>',	"GT"	},
	//{ BCDEHLMA + BDHPSW + BDHSP + REG,		H,	"H"	},
	{ UNARY  + UOP3  + OPR,				HIGH,	"HIGH"	},
	//{ BCDEHLMA + REG,				L,	"L"	},
	{ BINARY + RELAT + OPR,				LE,	"LE"	},
	{ UNARY  + UOP3  + OPR,				LOW,	"LOW"	},
	{ BINARY + RELAT + OPR,				'<',	"LT"	},
	//{ BCDEHLMA + REG,				M,	"M"	},
	{ BINARY + MULT  + OPR,				MOD,	"MOD"	},
	{ BINARY + RELAT + OPR,				NE,	"NE"	},
	{ UNARY  + UOP2  + OPR,				NOT,	"NOT"	},
	{ BINARY + LOG2  + OPR,				OR,	"OR"	},
	//{ BDHPSW + REG,				PSW,	"PSW"	},
	{ BINARY + MULT  + OPR,				SHL,	"SHL"	},
	{ BINARY + MULT  + OPR,				SHR,	"SHR"	},
	//{ BDHSP + REG,				SP,	"SP"	},
	{ BINARY + LOG2  + OPR,				XOR,	"XOR"	}
    };

    return bccsearch(oprtbl,oprtbl + (sizeof(oprtbl) / sizeof(OPCODE)),nam);
}

//static OPCODE *bccsearch(OPCODE *lo, OPCODE *hi, char *nam)
//
//{
//    SCRATCH int i;
//    SCRATCH OPCODE *chk;
//
//  for (;;) {
//	chk = lo + (hi - lo) / 2;
//	if (!(i = ustrcmp(chk -> oname,nam))) return chk;
//	if (chk == lo) return NULL;
//	if (i < 0) lo = chk;
//	else hi = chk;
//  }
//}

static OPCODE *bccsearch(OPCODE *lo, OPCODE *hi, char *nam)
//HRJ simple search to avoid sorting op code table.

{
	SCRATCH int i;
	SCRATCH OPCODE *chk;

	for (chk = lo; chk < hi; chk++) {
	   if (!(i = ustrcmp(chk -> oname,nam))) return chk;
	}
	return NULL;
}

static int ustrcmp(char *s, char *t)

{
    SCRATCH int i;

    while (!(i = toupper(*s++) - toupper(*t)) && *t++);
    return i;
}

/*  Buffer storage for line listing routine.  This allows the listing	*/
/*  output routines to do all operations without the main routine	*/
/*  having to fool with it.						*/

static FILE *list = NULL;

/*  Listing file open routine.  If a listing file is already open, a	*/
/*  warning occurs.  If the listing file doesn't open correctly, a	*/
/*  fatal error occurs.  If no listing file is open, all calls to	*/
/*  lputs() and lclose() have no effect.				*/

void lopen(char *nam)

{

    if (list) warning(TWOLST);
    else if (!(list = fopen(nam,"w"))) fatal_error(LSTOPEN);
    return;
}

/*  Listing file line output routine.  This routine processes the	*/
/*  source line saved by popc() and the output of the line assembler in	*/
/*  buffer obj into a line of the listing.  If the disk fills up, a	*/
/*  fatal error occurs.							*/

void lputs(void)
{
    SCRATCH int i, j;
    SCRATCH unsigned *o;
    void fatal_error(char *);

    if (list) {
	i = bytes;  o = obj;
	do {
	    fprintf(list,"%c  ",errcode);
	    if (listhex) {
		fprintf(list,"%04x  ",address);
		for (j = 4; j; --j) {
		    if (i) { --i;  ++address;  fprintf(list," %02x",*o++); }
		    else fprintf(list,"   ");
		}
	    }
	    else fprintf(list,"%18s","");
	    fprintf(list,"   %s",lline);  strcpy(lline,"\n");
	    check_page();
	    if (ferror(list)) fatal_error(DSKFULL);
	} while (listhex && i);
    }
    return;
}

/*  Listing file close routine.  The symbol table is appended to the	*/
/*  listing in alphabetic order by symbol name, and the listing file is	*/
/*  closed.  If the disk fills up, a fatal error occurs.		*/

static int col = 0;

void lerror(void)
{
	 if (errors && list) fprintf(list, "%d Error(s)\n",errors); //hrj
}

void lclose(void)
{

    if (list) {
	if (sroot) {
	    list_sym(sroot);
	    if (col) fprintf(list,"\n");
	}
	fprintf(list,"\f");
	if (ferror(list) || fclose(list) == EOF) fatal_error(DSKFULL);
    }
    return;
}

static void list_sym(SYMBOL *sp)

{

    if (sp) {
	list_sym(sp -> left);
	fprintf(list,"%04x  %-10s",sp -> valu,sp -> sname);
	if (col = ++col % SYMCOLS) fprintf(list,"    ");
	else {
	    fprintf(list,"\n");
	    if (sp -> right) check_page();
	}
	list_sym(sp -> right);
    }
    return;
}

static void check_page(void)
{
    if (pagelen && !--listleft) eject = TRUE;
    if (eject) {
	eject = FALSE;  listleft = pagelen;  fprintf(list,"\f");
	if (title[0]) { listleft -= 2;  fprintf(list,"%s\n\n",title); }
    }
    return;
}

/*  Buffer storage for hex output file.  This allows the hex file	*/
/*  output routines to do all of the required buffering and record	*/
/*  forming without the	main routine having to fool with it.		*/

static FILE *hex = NULL;
static unsigned cnt = 0;
static unsigned addr = 0;
static unsigned sum = 0;
static unsigned buf[HEXSIZE];

/*  Hex file open routine.  If a hex file is already open, a warning	*/
/*  occurs.  If the hex file doesn't open correctly, a fatal error	*/
/*  occurs.  If no hex file is open, all calls to hputc(), hseek(), and	*/
/*  hclose() have no effect.						*/

void hopen(char *nam)

{

    if (hex) warning(TWOHEX);
    else if (!(hex = fopen(nam,"w"))) fatal_error(HEXOPEN);
    return;
}

/*  Hex file write routine.  The data byte is appended to the current	*/
/*  record.  If the record fills up, it gets written to disk.  If the	*/
/*  disk fills up, a fatal error occurs.				*/

void hputc(unsigned c) // from hputc() HRJ

{

    if (hex) {
	buf[cnt++] = c;
	if (cnt == HEXSIZE) record(0);
    }
    return;
}

/*  Hex file address set routine.  The specified address becomes the	*/
/*  load address of the next record.  If a record is currently open,	*/
/*  it gets written to disk.  If the disk fills up, a fatal error	*/
/*  occurs.								*/

void hseek(unsigned a)

{

    if (hex) {
	if (cnt) record(0);
	addr = a;
    }
    return;
}

/*  Hex file close routine.  Any open record is written to disk, the	*/
/*  EOF record is added, and file is closed.  If the disk fills up, a	*/
/*  fatal error occurs.							*/

void hclose(void)
{

    if (hex) {
	   if (cnt) record(0);
	   record(1);
	   if (fclose(hex) == EOF) fatal_error(DSKFULL);
    }
    return;
}

static void record(unsigned typ)

{
    SCRATCH unsigned i;

	putc(':',hex);  putb(cnt);  putb(high(addr));
    putb(low(addr));  putb(typ);
    for (i = 0; i < cnt; ++i) putb(buf[i]);
    putb(low(0-sum));  putc('\n',hex); /* was (-sum) HRJ*/

    addr += cnt;  cnt = 0;

    if (ferror(hex)) fatal_error(DSKFULL);
    return;
}

static void putb(unsigned b)

{
    static char digit[] = "0123456789ABCDEF";

    putc(digit[b >> 4],hex);  putc(digit[b & 0x0f],hex);
    sum += b;  return;
}

/*  Error handler routine.  If the current error code is non-blank,	*/
/*  the error code is filled in and the	number of lines with errors	*/
/*  is adjusted.							*/

void error(char code)

{
    if (errcode == ' ') { errcode = code;  ++errors; }
    return;
}

/*  Fatal error handler routine.  A message gets printed on the stderr	*/
/*  device, and the program bombs.					*/

void fatal_error(char *msg)

{
    printf("Fatal Error -- %s\n",msg);
    exit(-1);
}

/*  Non-fatal error handler routine.  A message gets printed on the	*/
/*  stderr device, and the routine returns.				*/

void warning(char *msg)

{
    printf("Warning -- %s\n",msg);
    return;
}

